/* interp.c    interpreter i-code	9 Jan 97

	start project RZ				3 Apr 2001
	project A1						23 Oct 2002
	som-c change to dynamic array  	4 Jan 2004
	rewrite using big switch	  	13 Jan 2004
	new call/ret					22 Mar 2004
	add icSys 19 loadfile in som3	9 Nov 2004
	eliminate tail call				Feb 2005
 */

#include "compile.h"
#include "interp.h"

#define		PRIVATE 	static

/* run time environment
   define code, data, stack segments here

   system area  1 .. MAXSYS-1
   1..80   tokstring-som
*/

int M[MAXMEM];			/* data and code segment */
int SS[MAXSS];			/* stack segment */
int *CS = &M[MAXDS];	/* code segment */
int DS = MAXSYS;		/* current free data segment */

PRIVATE int sp,fp,ip;	/* current state of computation */
PRIVATE int opcode,data;/* current instruction */
PRIVATE int runflag;
// PRIVATE char name[80];	/* file name for syscall 19 */

/*
PRIVATE int traceflag = 0;

extern FILE *FO;
extern int lv;

PRIVATE void trace(void){
	int i, a, b;
	printf("%5d ",ip);
	prCode(ip);
	printf(" |");
	a = sp - 10;
	if( a < 0 ) a = 0;

	for(i=a; i<=sp; i++)
		if(i == fp)
			printf("[%d] ",SS[i]);
		else
			printf("%d ",SS[i]);
	printf("\n");
}

void dumpDS(void){
	int i;
	for(i=1; i<10; i++)	printf("%d:%d ",i,M[i]);
	printf("\n");
}
*/

PRIVATE void error(char *mess){
	printf("%s at ip %d\n",mess,ip);
	exit(0);
}

PRIVATE void exec(int op, int data);

void eval(int ref){
//	int limit = 200000;

	sp = 1;
	fp = 0;
	ip = ref;
	runflag = 1;
    while(runflag ) {
//		if( (limit--) < 0 ) break;
//		if (traceflag) trace();

		opcode = CS[ip] & 255;
		data = CS[ip] >> 8 ;
		if (opcode > EOC)
			error("undefined code");
		exec(opcode,data);
    }
//	dumpDS();
}

PRIVATE void push(int x) { sp++; SS[sp] = x; }
PRIVATE int pop(void){ int x; x = TOS; pop1; return x; }

// static alloc memory of size a
int xalloc(int a){
	int b;
	b = DS;
	if( DS + a >= MAXDS )
		seterror("alloc: out of memory");
	DS += a;
	return b;
}

PRIVATE void exec(int op, int data){
	int a, b, i, low, hi, var;

	switch(op){
	case icLit:
		push(data);
		ip++;
		break;
	case icLd:
		push( M[data]);
		ip++;
		break;
	case icSt:
		M[data] = TOS;
		pop1;
		ip++;
		break;
	case icPut:
		SS[fp-data] = pop();
		ip++;
		break;
	case icGet:
		push(SS[fp-data]);
		ip++;
		break;
	case icLdx:
		a = NEXT + TOS;
		pop1;
		TOS = M[a];
		ip++;
		break;
	case icStx:
		a = pop();
		b = NEXT + TOS;
		pop2;
		M[b] = a;
		ip++;
		break;
	case icArray:
		a = pop();
		push(xalloc(a));
		ip++;
		break;
	case icJmp:
		ip += data;
		break;
	case icJf:
		if( pop() == FALSE) ip += data;
		else ip++;
		break;
	case icJt:
		if( pop() != FALSE) ip += data;
		else ip++;
		break;
	case icCall:	/* a new smaller call in som4 */
		var = CS[data] >> 8;		// get lv
		if (sp + var + FUDGE > MAXSS)
			error("stack overflow");
		SS[sp+var] = fp;
		fp = sp + var;
		sp = fp + 1;
		SS[sp] = ip + 1;
		ip = data + 1;
		break;
	case icRet:
	// return TOS to caller if current stack is not empty
		ip = SS[fp+1];
		if ( sp > fp+1 ){
			a = TOS;
			sp = fp - data + 1;
			fp = SS[fp];
			SS[sp] = a;
		}else{
			sp = fp - data;
			fp = SS[fp];
		}
		break;
	case icAdd:
		a = NEXT + TOS;
		pop1;
		TOS = a;
		ip++;
		break;
	case icSub:
		a = NEXT - TOS;
		pop1;
		TOS = a;
		ip++;
		break;
	case icMul:
		a = NEXT * TOS;
		pop1;
		TOS = a;
		ip++;
		break;
	case icDiv:
		a = NEXT / TOS;
		pop1;
		TOS = a;
		ip++;
		break;
	case icBand:
		a = NEXT & TOS;
		pop1;
		TOS = a;
		ip++;
		break;
	case icBor:
		a = NEXT | TOS;
		pop1;
		TOS = a;
		ip++;
		break;
	case icBxor:
		a = NEXT ^ TOS;
		pop1;
		TOS = a;
		ip++;
		break;
	case icNot:
		a = TOS == FALSE;
		TOS = a;
		ip++;
		break;
	case icShl:
		a = NEXT << TOS;
		pop1;
		TOS = a;
		ip++;
		break;
	case icShr:
		a = NEXT >> TOS;
		pop1;
		TOS = a;
		ip++;
		break;
	case icEq:
		a = NEXT == TOS;
		pop1;
		TOS = a;
		ip++;
		break;
	case icNe:
		a = NEXT != TOS;
		pop1;
		TOS = a;
		ip++;
		break;
	case icLt:
		a = NEXT < TOS;
		pop1;
		TOS = a;
		ip++;
		break;
	case icLe:
		a = NEXT <= TOS;
		pop1;
		TOS = a;
		ip++;
		break;
	case icGt:
		a = NEXT > TOS;
		pop1;
		TOS = a;
		ip++;
		break;
	case icGe:
		a = NEXT >= TOS;
		pop1;
		TOS = a;
		ip++;
		break;
	case icInc:
		SS[fp-data]++;
		ip++;
		break;
	case icDec:
		SS[fp-data]--;
		ip++;
		break;
	case icCase:
		low = CS[ip+1] >> 8;
		hi = CS[ip+2] >> 8;
		a = pop();
		if( a < low || a > hi )
			ip = ip+3;
		else
			ip += a-low+4;
		break;
	case icSys:
		switch(data){
		case 1: printf("%d",pop()); break;  /* pr int */
		case 2: printf("%c",pop()); break;	/* pr char */
		case 3: push(getchar()); break;		/* getchar */
/*
		case 19:			// loadfile
			a = pop();		// file name in som-string
			strunpack(name,a);
			#ifdef VERBOSE
				printf("load %s\n",name);
			#endif
			loadfile(name);
			break;
*/
		default: seterror("undefined syscall");
		}
		ip++;
		break;

	case icEnd:
		runflag = 0;
		break;

	default:
		error("undefined opcode");
	}
}


