/* lex.c   lexical analyzer R1		15 Oct 97

	start project RZ				3 Apr 2001
	project A1c						27 Nov 2002
	use indentation for blocking	25 Oct 2003
	add hex                         1 Dec 2003
*/

#include "compile.h"

#define MAXFIN	100000	/* max input file size */
#define MAXLINE 1000	/* max line buf size */
#define maxcol  50		/* max col stk depth */
#define MAXLEX	50		/* lex state stk depth */

#define	cBLANK		10
#define	cLETTER     11
#define	cDIGIT      12
#define cQUOTE		13
#define cSPECIAL	14
#define cHEX		15

#define	PRIVATE	static

extern int token(void);	/* in token.c */

int tok; 				/* current token type */
char tokstring[MAXSTR];	/* current token string */
int line = 1, col = 0;	/* current position in source */
uchar CH;       		/* current char */

//PRIVATE char inbuf[MAXFIN];		/* input file buffer */
PRIVATE char *cp = NULL;
PRIVATE char chartype[256];			/* character type table */

PRIVATE int colstk[maxcol], colsp = 0;  /* column stack */
PRIVATE char lexstate = 'N';

typedef struct {					// state of lex
	char *cp;
	int tok;
	char tokstring[MAXSTR];
	int line;
} reclex;

PRIVATE reclex stacklex[MAXLEX];	// stack to save lex state
PRIVATE int splex = 0;				// stack pointer

PRIVATE char tokstr[][8] = {		/* print name of token */
	"*","/","-","+","=","==","&","|","^","%","!",
	"!=","<","<=","<<",">",">=",">>",":","(",")","[","]",
	"{","}","to","if","else","while","for","break","array",
	"case","enum","syscall"
};

void prtok(int tok){
	if( tok - tkSTAR > tkSYSCALL )	printf("unknown token");
	else printf("%s ",tokstr[tok-tkSTAR]);
}

void initCharType(void){
  int i;
  for(i=0; i<256; i++) chartype[i] = cBLANK;
  for(i=33; i<128; i++ ) chartype[i] = cSPECIAL;
  for(i='0'; i<='9'; i++ ) chartype[i] = cDIGIT;
  for(i='A'; i<='Z'; i++ ) chartype[i] = cLETTER;
  for(i='a'; i<='z'; i++ ) chartype[i] = cLETTER;
  chartype['_'] = cLETTER;
  chartype['"'] = cQUOTE;
  chartype[EOF_CHAR] = cSPECIAL;
  chartype['#'] = cHEX;
}

void getC(void){		// stop at EOF
  if ( CH != EOF_CHAR ) {
    CH  = *cp++;
	col++;
  }
}

PRIVATE void unget(void) {
	cp--;
	CH = *(cp-1);
	col--;
}

PRIVATE int isDigit(uchar c) {
	return (chartype[c] == cDIGIT);
}

int isLetterOrDigit(uchar c) {
  return chartype[c] == cLETTER || chartype[c] == cDIGIT;
}
/*
PRIVATE int isHex(uchar c){
	return (c >= '0' && c <= '9') ||
	  (c >= 'A' && c <= 'F') ||
	  (c >= 'a' && c <= 'f');
}

PRIVATE int htoi(char *s){
	char c;
	int v, h, i;
	h = 0;
	v = 0;
	i = 0;
	while( (c = *s) != 0 ){
		if( c >= '0' && c <= '9' ) v = (int)(c - '0');
		else if( c >= 'A' && c <= 'F' ) v = (int)(c - 'A' + 10);
		else if( c >= 'a' && c <= 'f' ) v = (int)(c - 'a' + 10);
		if( i == 0 && v > 7 )
			h = v - 16;			// sign extension
		else
			h = h * 16 + v;
		s++;
		i++;
	}
	return h;
}
*/

PRIVATE void skipblank(void){
  for(;;) {
    while( chartype[CH] == cBLANK ){
      if ( CH == NL_CHAR) {
		  line++;
		  col = 0;
	  }
      getC();
    }
    if( CH != '/' ) break;
    getC();
    if( CH != '/' ) { unget(); break; }
    while( CH != NL_CHAR && CH != EOF_CHAR) getC();
  }
}

PRIVATE void copyToken(char* a, char* b) {
  int len;
  len = b - a;
  if( len <= 0 || len > MAXSTR)
	  seterror("lex token length");
  strncpy(tokstring,a,len);
  tokstring[len] = 0;
}

void prToken(int tok){
	switch(tok){
	case tkIDEN:
	case tkNUMBER: printf("%s ",tokstring); break;
	case tkSTRING: printf("%c%s%c ",34,tokstring,34); break;
	case tkEOF:	   printf("eof "); break;
	case tkERROR:  printf("token Error "); break;
	default:	   prtok(tok);
	}
}

PRIVATE void pushcol(int c){
	if( colsp >= maxcol )
		seterror("lex col stack overflow");
	else {
		colsp++;
		colstk[colsp] = c;
	}
}

PRIVATE void popcol(void){
	if( colsp <= 0 )
		seterror("lex col stack underflow");
	else
		colsp--;
}

PRIVATE void lex1(void){
  int a;
  char *cp1;

  if( CH == EOF_CHAR ) {
    tok = tkEOF;
    return;
  }
  cp1 = cp - 1;
  a = token();				// check token
  if( a != 0 ){				// accept
    cp = cp1 + a;
    getC();
	return;
  }
  /* recogniser reject, reset char pointer */
  cp = cp1;
  getC();
  switch( chartype[CH] ) {
  case cLETTER :
    while( isLetterOrDigit(CH) ) getC();
    copyToken(cp1,cp-1);
    tok = tkIDEN;
    break;
  case cDIGIT :
    while( isDigit(CH) ) getC();
    copyToken(cp1,cp - 1);
    tok = tkNUMBER;
    break;
/*
  case cHEX :
	getC();     // skip #
	while( isHex(CH) ) getC();
	copyToken(cp1,cp - 1);
	tok = tkNUMBER;
	tokvalue = htoi(tokstring+1);
	break;
*/
  case cQUOTE :
    do {
      getC();
      if ( CH == NL_CHAR ) line++;
    }  while( CH != '"' && CH != EOF_CHAR);
    copyToken(cp1+1,cp-1);
    getC();
    tok = tkSTRING;
    break;

  default : tok = tkERROR;
  }
}

/* lexstate
 state description   condition  action    next
   N  normal  lex     oldline   out tok     N
   N                  nl & ==   out tok     N
   N                  nl & >    out { push  F
   N                  nl & <    out } pop   B
   F  forward         -         out oldtok  N
   B  backward        ==        out oldtok  N
   B                  <         out } pop   B
   transition occurs on call lex()
   eof is a special case
 */
void lex(void){
  int oldline;
  static int tokcol, oldtok;

  switch(lexstate){
  case 'N':  // normal
    oldline = line;
	skipblank();
  	tokcol = col;
    lex1();
//	prToken(tok);
	oldtok = tok;
    if( tok == tkEOF ) { // out block-end until match
      if( colstk[colsp] != 1 ){
        tok = tkBE;
        popcol();
      }
	  break;
	}

    if( line != oldline ){ // newline
      if( tokcol > colstk[colsp] ){
		tok = tkBB;
		pushcol(tokcol);
        lexstate = 'F';
	  }else if( tokcol < colstk[colsp] ){
		tok = tkBE;
		popcol();
		lexstate = 'B';
      }  // tokcol == colstk, just return tok
    }
	break;
  case 'F': // forward
    tok = oldtok;
    lexstate = 'N';
    break;
  case 'B':  // backward
    if( tokcol >= colstk[colsp] ){
	  tok = oldtok;
	  lexstate = 'N';
	}else if( tokcol < colstk[colsp] ){
	  tok = tkBE;
	  popcol();
	}
	break;
  }
}

PRIVATE char *getbuf(int size){
	char *c;
	c = malloc(size);
	if( c == NULL ){
		printf("malloc fail at read\n");
		exit(-1);
	}
	return c;
}

void readinfile(char *fname) {
	FILE *fp;
	int n;
	char *cbuf;
	fp = fopen(fname,"rt");
	if( fp == NULL ) {
		printf("cannot open : %s\n",fname);
		exit(-1);
	}
	cbuf = getbuf(MAXFIN);
	n = fread(cbuf,1,MAXFIN-2,fp);
	fclose(fp);
	cbuf[n] = EOF_CHAR;
	cbuf[n+1] = 0;
	savelex();
	cp = cbuf;
	line = 1;
}

// read one line from stdin, skip blank line
void reader(void){
	int n;
	char *cbuf;
	cbuf = getbuf(MAXLINE);
	do{
		printf(">");
		fgets(cbuf,MAXLINE-2,stdin);
		n = strspn(cbuf," \t\n");   // check blank
	}while( cbuf[n] == '\0');
	n = strlen(cbuf);
	cbuf[n] = EOF_CHAR;
	cbuf[n+1] = 0;
	savelex();
	cp = cbuf;
	line = 1;
}

void freebuf(void){
	free(cp);
	restorelex();
}

void startlex(void){	// must readinfile first
	CH = *cp;			// clear old CH
	col = 1;
	getC();
//	tokcol = 1;
	pushcol(1);
//	lex();
}

void savelex(void){
	splex++;
	if(splex > MAXLEX)
		seterror("recursive load overflow");
	stacklex[splex].cp = cp;
	stacklex[splex].tok = tok;
	strcpy(stacklex[splex].tokstring,tokstring);
	stacklex[splex].line = line;
}

void restorelex(void){
	if(splex <= 0)
		seterror("recursive load underflow");
	cp = stacklex[splex].cp;
	tok = stacklex[splex].tok;
	strcpy(tokstring,stacklex[splex].tokstring);
	line = stacklex[splex].line;
	splex--;

	if(cp != 0){
		CH = *cp;		// start lex
		getC();
	}
}


